;*=====================================================================*/
;*    serrano/prgm/project/bigloo/bmacs/bdb-display.el                 */
;*    -------------------------------------------------------------    */
;*    Author      :  Manuel Serrano                                    */
;*    Creation    :  Thu Apr 30 10:24:19 1998                          */
;*    Last change :  Sun Oct 25 07:50:56 1998 (serrano)                */
;*    -------------------------------------------------------------    */
;*    This file implement the bdb display printing.                    */
;*=====================================================================*/

;*---------------------------------------------------------------------*/
;*    The module                                                       */
;*---------------------------------------------------------------------*/
(provide 'bdb-display)
(require 'bdb-config)
(require 'bdb)

;*---------------------------------------------------------------------*/
;*    Global control variables.                                        */
;*---------------------------------------------------------------------*/
;; The completion process filter indicates when it is finished.
(defvar bdb-display-in-progress)

;; Since output may arrive in fragments we accumulate partials strings here.
(defvar bdb-display-string)

;; this variable holds a state: are we currently inquirying for displays
(defvar bdb-display-command-armed-p nil)

;; the buffer for C display
(defvar bdb-display-buffer nil)

;; which display to display (Scheme or C)?
(defvar bdb-display 'scheme)

;; display in frame or window
(defvar bdb-display-in-frame-p t)

;*---------------------------------------------------------------------*/
;*    bdb-display-hook ...                                             */
;*---------------------------------------------------------------------*/
(defun bdb-display-hook (input)
  (if (and (not bdb-display-command-armed-p)
	   (or (not (stringp input))
	       ;; we invoke this hook on:
	       ;; next, step, continue, until, call, finish, run,
	       ;; return and breturn.
	       ;; it is important not to re-call the hook on `frame...'
	       (string-match "^[nscru]\\|^b[sdrncuf]\\|^disp\\|^fi" input)))
      (progn
	(if (not (bufferp bdb-display-buffer))
	    (setq bdb-display-buffer (bdb-make-display-buffer "*Bdb display*"))
	  (bdb-popup-display bdb-display-buffer))
	(bdb-display-command bdb-display-buffer))))

;*---------------------------------------------------------------------*/
;*    bdb-display-command ...                                          */
;*    -------------------------------------------------------------    */
;*    This command refresh the display tracing.                        */
;*---------------------------------------------------------------------*/
(defun bdb-display-command (buffer)
  "Perform completion on the BDB command preceding point.
This is implemented using the BDB `display' command." 
  (interactive)
  (if (not bdb-display-command-armed-p)
      (progn
	;; we mark that display is armed not to re-enter this hook
	(setq bdb-display-command-armed-p t)
	;; we wait for comint to be ready
	(bdb-wait-for-comint "bdb-display-command")
	;; Temporarily install our filter function.
	(let ((bdb-marker-filter 'bdb-display-filter))
	  ;; Issue the command to BDB.
	  (bdb-info-display-call)
	  (setq bdb-display-in-progress t)
	  (setq bdb-display-string "")
	  ;; Slurp the output.
	  (while bdb-display-in-progress
	    (if (not (accept-process-output
		      (get-buffer-process bdb-comint-buffer)
		      bdb-wait-output-timeout ))
		(progn
		  (message "Display timeout...")
		  (setq bdb-display-in-progress nil)
		  (setq bdb-display-string nil)))))
	;; at this point bdb-display-string is the string of all
	;; the display, we have to parse in order to find which
	;; display are still actives.
	(if (stringp bdb-display-string)
	    (bdb-display-display buffer bdb-display-string))
	(setq bdb-display-command-armed-p nil)))
  "")

;*---------------------------------------------------------------------*/
;*    bdb-info-display-call ...                                        */
;*    -------------------------------------------------------------    */
;*    Sending a bdb command.                                           */
;*---------------------------------------------------------------------*/
(defun bdb-info-display-call ()
  (interactive)
  (let ((command "bdisplay\n")
	(proc (get-buffer-process bdb-comint-buffer)))
    (or proc (error "Current buffer has no process"))
    ;; Arrange for the current prompt to get deleted.
    (save-excursion
      (set-buffer bdb-comint-buffer)
      (goto-char (process-mark proc))
      (while (not (looking-at comint-prompt-regexp))
	(goto-char (process-mark proc))
	(beginning-of-line))
      (process-send-string proc command))))

;*---------------------------------------------------------------------*/
;*    bdb-display-filter ...                                           */
;*    -------------------------------------------------------------    */
;*    The completion process filter is installed temporarily to slurp  */
;*    the output of BDB up to the next prompt and build the completion */
;*    list.                                                            */
;*---------------------------------------------------------------------*/
(defun bdb-display-filter (string)
  (setq bdb-marker-acc "")
  (setq bdb-display-string (concat bdb-display-string string))
  (if (string-match comint-prompt-regexp string)
      (setq bdb-display-in-progress nil))
  "")

;*---------------------------------------------------------------------*/
;*    The display mode map                                             */
;*---------------------------------------------------------------------*/
(defvar bdb-display-mode-map (make-sparse-keymap))
(define-key bdb-display-mode-map bdb-mouse-binding
  'bdb-display-mode-menu)
(defvar bdb-display-mouse-map (make-sparse-keymap))
(define-key bdb-display-mouse-map [(button2)]
  'bdb-display-menu)
(define-key bdb-display-mode-map "\C-x\C-c" 'bdb-display-quit)

;*---------------------------------------------------------------------*/
;*    bdb-display-display ...                                          */
;*    -------------------------------------------------------------    */
;*    Display a new display in a separate buffer.                      */
;*---------------------------------------------------------------------*/
(defun bdb-display-display (buffer display)
  (save-excursion
    (set-buffer buffer)
    (let ((buffer-read-only nil))
      (erase-buffer)
      (insert display)
      (save-excursion
	(goto-char (point-max))
	(re-search-backward comint-prompt-regexp))
      ;; we remove the bdb prompt
      (delete-region (match-beginning 0) (match-end 0))
      ;; we insert extent in all text
      (save-excursion
	(goto-char (point-min))
	(let ((keep t))
	  (while keep
	    (beginning-of-line)
	    (let ((start (point)))
	      (if (looking-at "^[0-9]+")
		  ;; this line is a frame line
		  (let ((extent (make-extent 1 1 buffer))
			(end    (progn (end-of-line) (point))))
		    (set-extent-property extent 'mouse-face 'highlight)
		    (set-extent-property extent 'keymap bdb-display-mouse-map)
		    (set-extent-endpoints extent start end)
		    (if (< end (point-max))
			(next-line 1)
		      (setq keep nil)))
		(progn
		  (end-of-line)
		  (if (< (point) (point-max))
		      (next-line 1)
		    (setq keep nil)))))))))))
		
;*---------------------------------------------------------------------*/
;*    bdb-make-display-buffer ...                                      */
;*---------------------------------------------------------------------*/
(defun bdb-make-display-buffer (name)
  (if bdb-display-in-frame-p
      (bdb-make-display-frame-buffer name)
    (bdb-make-display-window-buffer name)))

;*---------------------------------------------------------------------*/
;*    bdb-make-display-frame-buffer ...                                */
;*---------------------------------------------------------------------*/
(defun bdb-make-display-frame-buffer (name)
  (let ((height bdb-display-height))
    (let* ((default-frame-alist (cons 'height
				      (cons height
					    default-frame-alist)))
	   (buffer (switch-to-buffer-other-frame name)))
      (set-buffer buffer)
      (bdb-display-mode)
      (current-buffer))))

;*---------------------------------------------------------------------*/
;*    bdb-make-display-window-buffer ...                               */
;*---------------------------------------------------------------------*/
(defun bdb-make-display-window-buffer (name)
  (set-buffer bdb-comint-buffer)
  (let* ((window (selected-window))
	 (height (window-height window)))
    (split-window window)
    (let ((buffer (switch-to-buffer-other-window name)))
      (set-buffer buffer)
      (bdb-display-mode)
      (current-buffer))))

;*---------------------------------------------------------------------*/
;*    bdb-popup-display ...                                            */
;*    -------------------------------------------------------------    */
;*    If the frame displaying a display buffer has been killed, we     */
;*    popup a fresh one.                                               */
;*---------------------------------------------------------------------*/
(defun bdb-popup-display (buffer)
  (let ((window (get-buffer-window buffer 0)))
    (if (not (windowp window))
	(bdb-make-display-buffer (buffer-name buffer)))))
	
;*---------------------------------------------------------------------*/
;*    bdb-display-mode ...                                             */
;*---------------------------------------------------------------------*/
(defun bdb-display-mode ()
  "Major mode for bdb display.

\\{bdb-display-mode-map}"
  (setq major-mode 'bdb-display-mode)
  (setq mode-name "Bdb Display")
  (setq buffer-read-only t)
  (setq comint-prompt-regexp bdb-prompt-regexp)
  (setq paragraph-start comint-prompt-regexp)
  (use-local-map bdb-display-mode-map)
  (suppress-keymap bdb-display-mode-map)
  (font-lock-mode t)
  (setq font-lock-keywords-case-fold-search t)
  (setq font-lock-keywords (list
			    (list "^[0-9]+:\\s-+\\(\\S-+\\).+(!)"
				  1
				  bdb-c-frame-face)
			    (list "^[0-9]+:\\s-+\\(\\S-+\\).+(-)"
				  1
				  bdb-scheme-frame-face))))
  
;*---------------------------------------------------------------------*/
;*    bdb-display-menu ...                                             */
;*---------------------------------------------------------------------*/
(defun bdb-display-menu (event)
  (interactive "e")
  (let (selection)
    (save-excursion
      (set-buffer (window-buffer (event-window event)))
      (save-excursion
	(goto-char (event-point event))
	(message "point: %S" (point))
	(setq selection (bdb-get-display-number))))
    (popup-menu (list (concat "display " selection)
		      (vector "disable"
			      (list 'bdb-remote-call
				    (concat "disable display " selection))
			      t)
		      (vector "undisplay"
			      (list 'bdb-remote-call
				    (concat "undisplay " selection))
			      t)))))

;*---------------------------------------------------------------------*/
;*    bdb-get-display-number ...                                       */
;*---------------------------------------------------------------------*/
(defun bdb-get-display-number ()
  (save-excursion
    (let ((pos (re-search-backward "^[0-9]+" nil t)))
      (or (and pos (buffer-substring (match-beginning 0) (match-end 0)))
	  "0"))))

;*---------------------------------------------------------------------*/
;*    bdb-display-mode-menu ...                                        */
;*---------------------------------------------------------------------*/
(defun bdb-display-mode-menu (event)
  (interactive "e")
  (popup-menu (list "bdb display"
		    (vector (format "Set display height (current: %d)"
				    bdb-display-height)
			    'bdb-display-set-height
			    t)
		    (vector "Current height + 3"
			    '(bdb-display-set-height (+ bdb-display-height 3))
			    t)
		    (vector "Current height - 3"
			    '(bdb-display-set-height (- bdb-display-height 3))
			    t)
		    "--:shadowEtchedOut"
		    (vector "Close display frame display"
			    '(bdb-display-quit)
			    t))))

;*---------------------------------------------------------------------*/
;*    bdb-display-set-height ...                                       */
;*---------------------------------------------------------------------*/
(defun bdb-display-set-height (height)
  (interactive "nHeight: ")
  (if (> height 0)
      (progn
	(setq bdb-display-height height)
	(set-frame-height (selected-frame) bdb-display-height))))

;*---------------------------------------------------------------------*/
;*    bdb-display-start ...                                            */
;*---------------------------------------------------------------------*/
(defun bdb-display-start (framep)
  (interactive)
  (setq bdb-display-in-frame-p framep)
  (if (not (bdb-installed-hook-p 'bdb-display-hook))
      (progn
	(bdb-add-send-input-hook 'bdb-display-hook)
	(bdb-display-hook t))))

;*---------------------------------------------------------------------*/
;*    bdb-display-quit ...                                             */
;*---------------------------------------------------------------------*/
(defun bdb-display-quit ()
  (interactive)
  (bdb-delete-window-or-frame bdb-display-buffer)
  (bdb-remove-send-input-hook 'bdb-display-hook))

;*---------------------------------------------------------------------*/
;*    bdb-display-toggle ...                                           */
;*---------------------------------------------------------------------*/
(defun bdb-display-toggle ()
  (interactive)
  (if (not (bdb-installed-hook-p 'bdb-display-hook))
      (bdb-display-start t)
    (bdb-display-quit)))
