;*=====================================================================*/
;*    serrano/prgm/project/bigloo/bmacs/bdb.el                         */
;*    -------------------------------------------------------------    */
;*    Author      :  Manuel Serrano                                    */
;*    Creation    :  Sun Apr 19 15:16:46 1998                          */
;*    Last change :  Sun Oct 25 08:45:17 1998 (serrano)                */
;*    -------------------------------------------------------------    */
;*    The bdb function.                                                */
;*=====================================================================*/

;*---------------------------------------------------------------------*/
;*    The module                                                       */
;*---------------------------------------------------------------------*/
(provide 'bdb)
(require 'comint)
(require 'bdb-autoload)
(require 'bdb-process)
(require 'bdb-filter)
(require 'bdb-complete)
(require 'bdb-mode)
(require 'bdb-breakpoint)

(setq debug-on-error t)
(setq debug-on-quit t)

;*---------------------------------------------------------------------*/
;*    Comint global control.                                           */
;*---------------------------------------------------------------------*/
(defvar bdb-comint-buffer nil)

;*---------------------------------------------------------------------*/
;*    bdb-history ...                                                  */
;*---------------------------------------------------------------------*/
(defvar bdb-history '())

;*---------------------------------------------------------------------*/
;*    bdb ...                                                          */
;*---------------------------------------------------------------------*/
(defun bdb (command-line)
  "Run bdb on program FILE in buffer *gud-FILE*.
The directory containing FILE becomes the initial working directory
and source-file directory for your debugger."
  (interactive
   (list (let ((make-entry (bee-makefile-fetch-binary-entry)))
	   (if (stringp make-entry)
	       (concat bdb-binary " " make-entry)
	     (read-from-minibuffer "Run bdb (like this): "
				   (if (consp bdb-history)
				       (car bdb-history)
				     (concat bdb-binary " "))
				   bdb-minibuffer-local-map nil
				   '(bdb-history . 1))))))
  (setq bdb-history (cons command-line bdb-history))
  ;; we split the command line in list of words
  (let ((word-list (bdb-string->list command-line)))
    (if (= (length word-list) 2)
	(let* ((binary  (car word-list))
	       (a.out   (cadr word-list))
	       (bdbname (concat "bdb-" a.out))
	       (bufname (concat "*" bdbname "*")))
	  (setq bdb-binary binary)
	  ;; we create a *bdb* buffer
	  (switch-to-buffer-other-frame bufname)
	  ;; we set this buffer as the comint buffer
	  (setq bdb-comint-buffer (current-buffer))
	  ;; we start comint
	  (setq bdb-marker-filter 'bdb-default-marker-filter)
	  (make-comint bdbname binary nil "--emacs" a.out)
	  ;; the process filter and sentinel
	  (process-kill-without-query (get-buffer-process bdb-comint-buffer))
	  (set-process-filter (get-buffer-process bdb-comint-buffer)
			      'bdb-filter)
	  (set-process-sentinel (get-buffer-process bdb-comint-buffer)
				'bdb-sentinel)
	  ;; the output face
	  (make-face 'bdb-output-face)
	  (set-face-foreground 'bdb-output-face bdb-output-color)
	  ;; we prepare breakpoint tracing
	  (bdb-add-send-input-hook 'bdb-breakpoint-hook)
	  ;; we start the bdb-mode
	  (bdb-mode))
      (error (concat "Illegal bdb invokation: " command-line)))))

;*---------------------------------------------------------------------*/
;*    bdb-other-frame ...                                              */
;*---------------------------------------------------------------------*/
(defun bdb-other-frame ()
  (interactive)
  (message "Spawning Bdb...")
  (let ((buffer (current-buffer)))
    (if (bufferp bdb-comint-buffer)
	(let ((pop-up-frames t))
	  (display-buffer bdb-comint-buffer))
      (call-interactively 'bdb))))

;*---------------------------------------------------------------------*/
;*    bdb-string->list ...                                             */
;*    -------------------------------------------------------------    */
;*    This function split a string into a list of string. Space or     */
;*    tab is used as a word delimiter.                                 */
;*---------------------------------------------------------------------*/
(defun bdb-string->list (string)
  (let ((i 0) (beg 0)
	(len (length string))
	(words nil))
    (while (< i len)
      (if (memq (aref string i) '(?\t ? ))
	  (progn
	    (setq words (cons (substring string beg i) words))
	    (setq beg (1+ i))
	    (while (and (< beg len) (memq (aref string beg) '(?\t ? )))
	      (setq beg (1+ beg)))
	    (setq i (1+ beg)))
	(setq i (1+ i))))
    (if (< beg len)
	(setq words (cons (substring string beg) words)))
    (nreverse words)))

;*---------------------------------------------------------------------*/
;*    bdb-send-input-hooks ...                                         */
;*---------------------------------------------------------------------*/
(defvar bdb-send-input-hooks '())

;*---------------------------------------------------------------------*/
;*    bdb-add-send-input-hook ...                                      */
;*---------------------------------------------------------------------*/
(defun bdb-add-send-input-hook (hook)
  (setq bdb-send-input-hooks (cons hook bdb-send-input-hooks)))

;*---------------------------------------------------------------------*/
;*    bdb-remove-send-input-hook ...                                   */
;*---------------------------------------------------------------------*/
(defun bdb-remove-send-input-hook (hook)
  (setq bdb-send-input-hooks (delq hook bdb-send-input-hooks)))

;*---------------------------------------------------------------------*/
;*    bdb-installed-hook-p ...                                         */
;*    -------------------------------------------------------------    */
;*    Is a hook all set?                                               */
;*---------------------------------------------------------------------*/
(defun bdb-installed-hook-p (hook)
  (memq hook bdb-send-input-hooks))

;*---------------------------------------------------------------------*/
;*    bdb-remote-call ...                                              */
;*---------------------------------------------------------------------*/
(defun bdb-remote-call (command)
  (if bdb-verbose-remote
      (bdb-verbose-remote-call command)
    (bdb-silent-remote-call command)))

;*---------------------------------------------------------------------*/
;*    bdb-verbose-remote-call ...                                      */
;*    -------------------------------------------------------------    */
;*    This function is used for sending a bdb call from a remote       */
;*    buffer.                                                          */
;*---------------------------------------------------------------------*/
(defun bdb-verbose-remote-call (command)
  (if (not (bufferp bdb-comint-buffer))
      (error "No bdb running...")
    (progn
      (set-buffer bdb-comint-buffer)
      (let ((proc (get-buffer-process bdb-comint-buffer)))
	(goto-char (process-mark proc))
	(delete-region (point) (point-max))
	(insert command)
	(bdb-send-input)
	(goto-char (point-max))))))

;*---------------------------------------------------------------------*/
;*    bdb-silent-remote-call ...                                       */
;*---------------------------------------------------------------------*/
(defun bdb-silent-remote-call (command)
  (message "Command: %s" command)
  (sit-for 0)
  (bdb-basic-call command))

;*---------------------------------------------------------------------*/
;*    bdb-basic-call ...                                               */
;*    -------------------------------------------------------------    */
;*    Sending a bdb command.                                           */
;*---------------------------------------------------------------------*/
(defun bdb-basic-call (command)
  (interactive)
  (bdb-console-log "bdb-basic-call" command 'font-lock-function-name-face)
  (let ((command (concat command "\n"))
	(proc (get-buffer-process bdb-comint-buffer)))
    (or proc (error "Current buffer has no process"))
    ;; Arrange for the current prompt to get deleted.
    (save-excursion
      (set-buffer bdb-comint-buffer)
      (goto-char (process-mark proc))
      (beginning-of-line)
      (if (looking-at comint-prompt-regexp)
	  (set-marker bdb-delete-prompt-marker (point))))
    (process-send-string proc command)))

;*---------------------------------------------------------------------*/
;*    bdb-send-input ...                                               */
;*---------------------------------------------------------------------*/
(defun bdb-send-input ()
  (interactive)
  (let* ((proc  (get-buffer-process bdb-comint-buffer))
	 (pmark (process-mark proc))
	 (intxt (if (>= (point) (marker-position pmark))
		    (progn (if comint-eol-on-send (end-of-line))
			   (buffer-substring pmark (point)))
		  nil)))
    (if (stringp intxt)
	(bdb-console-log "bdb-send-input" intxt 'font-lock-type-face)
      (bdb-console-log "bdb-send-input" 'nil 'font-lock-string-face))
    (comint-send-input)
    (mapcar (lambda (hook) (funcall hook intxt)) bdb-send-input-hooks)))

;*---------------------------------------------------------------------*/
;*    Global control variables.                                        */
;*---------------------------------------------------------------------*/
;; The completion process filter indicates when it is finished.
(defvar bdb-no-output-in-progress)

;; Since output may arrive in fragments we accumulate partials strings here.
(defvar bdb-no-output-string)

;; this variable holds a state: are we currently inquirying for no-outputs
(defvar bdb-no-output-command-armed-p nil)

;*---------------------------------------------------------------------*/
;*    bdb-no-output-call ...                                           */
;*---------------------------------------------------------------------*/
(defun bdb-no-output-call (string)
  "Perform completion on the BDB command preceding point.
This is implemented using the BDB `no-output' command." 
  (interactive)
  (bdb-console-log "bdb-no-output-call" string 'font-lock-doc-string-face)
  (if (not bdb-no-output-command-armed-p)
      (progn
	;; we wait for comint to be ready
	(bdb-wait-for-comint "bdb-no-output-call")
	;; we mark that no-output is armed not to re-enter this hook
	(setq bdb-no-output-command-armed-p t)
	;; Temporarily install our filter function.
	(let ((bdb-marker-filter 'bdb-no-output-filter))
	  ;; Issue the command to BDB.
	  (bdb-info-no-output-call string)
	  (setq bdb-no-output-in-progress t)
	  (setq bdb-no-output-string "")
	  ;; Slurp the output.
	  (while bdb-no-output-in-progress
	    (accept-process-output (get-buffer-process bdb-comint-buffer))))
	;; at this point bdb-no-output-string is the string of all
	;; the no-output, we have to parse in order to find which
	;; no-output are still actives.
	(setq bdb-no-output-command-armed-p nil)
	bdb-no-output-string)
    ""))

;*---------------------------------------------------------------------*/
;*    bdb-info-no-output-call ...                                      */
;*    -------------------------------------------------------------    */
;*    Sending a bdb command.                                           */
;*---------------------------------------------------------------------*/
(defun bdb-info-no-output-call (string)
  (interactive)
  (let ((command (concat string "\n"))
	(proc (get-buffer-process bdb-comint-buffer)))
    (or proc (error "Current buffer has no process"))
    ;; Arrange for the current prompt to get deleted.
    (save-excursion
      (set-buffer bdb-comint-buffer)
      (goto-char (process-mark proc))
      (while (not (looking-at comint-prompt-regexp))
	(goto-char (process-mark proc))
	(beginning-of-line))
      (process-send-string proc command))))

;*---------------------------------------------------------------------*/
;*    bdb-no-output-filter ...                                         */
;*    -------------------------------------------------------------    */
;*    The completion process filter is installed temporarily to slurp  */
;*    the output of BDB up to the next prompt and build the completion */
;*    list.                                                            */
;*---------------------------------------------------------------------*/
(defun bdb-no-output-filter (string)
  (setq bdb-marker-acc "")
  (setq bdb-no-output-string (concat bdb-no-output-string string))
  (if (string-match comint-prompt-regexp string)
      (setq bdb-no-output-in-progress nil))
  "")

;*---------------------------------------------------------------------*/
;*    bdb-find-file ...                                                */
;*    -------------------------------------------------------------    */
;*    Find a file for bdb. Here we can add special keymap if we        */
;*    whish to.                                                        */
;*---------------------------------------------------------------------*/
(defun bdb-find-file (f)
  (interactive "Ffile: ")
  (if (file-exists-p f)
      (save-excursion
	(let ((buffer (find-file-noselect f)))
	  (if (bufferp buffer)
	      (bdb-connect-buffer buffer))
	  buffer))
    (let ((lst   (buffer-list))
	  (match (concat f "$"))
	  (buf   '()))
      (while (consp lst)
	(let* ((b     (car lst))
	       (bname (buffer-file-name b)))
	  (if (and (stringp bname) (string-match match bname))
	      ;; we have found the buffer
	      (progn
		(bdb-connect-buffer buf)
		(setq buf b)
		(setq lst '()))
	    (setq lst (cdr lst)))))
      buf)))

;*---------------------------------------------------------------------*/
;*    bdb-refresh ...                                                  */
;*---------------------------------------------------------------------*/
(defun bdb-refresh (&optional following)
  (interactive (list current-prefix-arg))
  (let ((bdb-mode nil))
    (recenter following)
    (mapcar (lambda (hook) (funcall hook nil)) bdb-send-input-hooks)))


;*---------------------------------------------------------------------*/
;*    bdb-comint-ready-p ...                                           */
;*    -------------------------------------------------------------    */
;*    This predicate returns T if and only if comint is ready for      */
;*    a new command.                                                   */
;*---------------------------------------------------------------------*/
(defun bdb-comint-ready-p ()
  (save-excursion
    (save-restriction
      (widen)
      (let ((proc (get-buffer-process bdb-comint-buffer)))
	(set-buffer bdb-comint-buffer)
	(goto-char (point-max))
	(let ((bound (point)))
	  (beginning-of-line)
	  (re-search-forward (concat comint-prompt-regexp "[ \t\n]*")
			     bound t))))))

;*---------------------------------------------------------------------*/
;*    bdb-wait-for-comint ...                                          */
;*    -------------------------------------------------------------    */
;*    We sit for the prompt is printed. This function is called        */
;*    by the send hooks.                                               */
;*---------------------------------------------------------------------*/
(defun bdb-wait-for-comint (from)
  (if (not bdb-comint-buffer)
      (error "No bdb running")
    (save-excursion
      (save-restriction
	(widen)
	(let ((proc  (get-buffer-process bdb-comint-buffer))
	      (count 0))
	  (set-buffer bdb-comint-buffer)
	  (accept-process-output)
	  (while (progn
		   (goto-char (point-max))
		   (let ((bound (point)))
		     (beginning-of-line)
		     (not (re-search-forward (concat
					      comint-prompt-regexp
					      "[ \t\n]*")
					     bound t))))
	    (setq count (+ 1 count))
	    (message "waiting for `%s' (count: %d max: %d substr: %S)"
		     from
		     count
		     (point-max)
		     (buffer-substring (point) (progn (end-of-line) (point))))
	    (sit-for 0.1 t)
	    (set-buffer bdb-comint-buffer)))))))

;*---------------------------------------------------------------------*/
;*    bdb-clear-window ...                                             */
;*---------------------------------------------------------------------*/
(defun bdb-clear-window ()
  (interactive)
  (set-buffer bdb-comint-buffer)
  (let ((proc (get-buffer-process bdb-comint-buffer)))
    (goto-char (process-mark proc))
    (beginning-of-line)
    (delete-region (point-min) (point))
    (goto-char (point-max))))

;*---------------------------------------------------------------------*/
;*    bdb-clear-line ...                                               */
;*---------------------------------------------------------------------*/
(defun bdb-clear-line ()
  (interactive)
  (set-buffer bdb-comint-buffer)
  (let ((proc (get-buffer-process bdb-comint-buffer)))
    (goto-char (process-mark proc))
    (kill-line 1)
    (goto-char (point-max))))

;*---------------------------------------------------------------------*/
;*    bdb-line-number ...                                              */
;*    -------------------------------------------------------------    */
;*    This function returns the point line number.                     */
;*---------------------------------------------------------------------*/
(defun bdb-line-number (buffer pos)
  (save-restriction
    (set-buffer buffer)
    (let (start)
      (save-excursion
	(save-restriction
	  (goto-char (point-min))
	  (widen)
	  (beginning-of-line)
	  (setq start (point))
	  (goto-char pos)
	  (beginning-of-line)
	  (if (/= start 1)
	      (1+ (count-lines 1 (point)))
	    (1+ (count-lines 1 (point)))))))))

;*---------------------------------------------------------------------*/
;*    bdb-delete-window-or-frame ...                                   */
;*    -------------------------------------------------------------    */
;*    If a buffer is the only one displayed in a frame, this function  */
;*    will delete the frame. Otherwise, the window displaying the      */
;*    buffer will be deleted.                                          */
;*---------------------------------------------------------------------*/
(defun bdb-delete-window-or-frame (buffer)
  (let ((win (get-buffer-window buffer t)))
    (if (windowp win)
	(let ((frame (window-frame win)))
	  (if (framep frame)
	      (save-excursion
		(select-window win)
		(if (one-window-p t)
		    (delete-frame frame)
		  (delete-window win)))
	    (delete-window win))))))

;*---------------------------------------------------------------------*/
;*    bdb-show ...                                                     */
;*---------------------------------------------------------------------*/
(defun bdb-show ()
  (interactive)
  (if (memq 'local bdb-show)
      (if (not (bdb-installed-hook-p 'bdb-locals-hook))
	  (bdb-locals-start nil)
	(bdb-locals-quit)))
  (if (memq 'args bdb-show)
      (if (not (bdb-installed-hook-p 'bdb-args-hook))
	  (bdb-args-start nil)
	(bdb-args-quit)))
  (if (memq 'display bdb-show)
      (if (not (bdb-installed-hook-p 'bdb-display-hook))
	  (bdb-display-start nil)
	(bdb-display-quit)))
  (if (memq 'stack bdb-show)
      (if (not (bdb-installed-hook-p 'bdb-stack-hook))
	  (bdb-stack-start nil)
	(bdb-stack-quit)))
  (let ((window (get-buffer-window bdb-comint-buffer)))
    (if (windowp window)
	(select-window window))))
