;*=====================================================================*/
;*    serrano/prgm/project/bigloo/bmacs/bee-docline.el                 */
;*    -------------------------------------------------------------    */
;*    Author      :  Manuel Serrano                                    */
;*    Creation    :  Mon Jun  1 07:58:46 1998                          */
;*    Last change :  Tue Oct 13 08:13:07 1998 (serrano)                */
;*    -------------------------------------------------------------    */
;*    The Bee online documentation. This file uses `xinfo' for         */
;*    displaying Info files.                                           */
;*=====================================================================*/

;*---------------------------------------------------------------------*/
;*    The module                                                       */
;*---------------------------------------------------------------------*/
(provide 'bee-docline)
(require 'bee-tools)
(require 'info)

;*---------------------------------------------------------------------*/
;*    bee-docline-init ...                                             */
;*    -------------------------------------------------------------    */
;*    We look if there is an existing                                  */
;*      bee-root-directory/info                                        */
;*    directory. If such exists, we add it to the search path for      */
;*    xinfo.                                                           */
;*---------------------------------------------------------------------*/
(defun bee-docline-init ()
  (if (stringp bee-root-directory)
      (let ((info (concat bee-root-directory "/info")))
	(if (and (file-exists-p info) (file-directory-p info))
	    (progn
	      (make-variable-buffer-local 'Info-directory-list)
	      (setq Info-directory-list (cons info Info-directory-list)))))))

;*---------------------------------------------------------------------*/
;*    bee-bigloo-info-file ...                                         */
;*---------------------------------------------------------------------*/
(defvar bee-bigloo-info-file
  "bigloo.info"
  "The name for the Bigloo info file.")

;*---------------------------------------------------------------------*/
;*    bee-info-file-list ...                                           */
;*    -------------------------------------------------------------    */
;*    The list of Info files to be search for documentation. This      */
;*    variable is buffer local an may contain extra info file that     */
;*    are project specific. For instance, the Bigloo compiler may      */
;*    contain its own extra info file. When searching for a function   */
;*    definition ref all the info file are searched.                   */
;*---------------------------------------------------------------------*/
(defvar bee-info-file-list
  (cons bee-bigloo-info-file '("r5rs.info"))
  "List of info files scanned when searching a function definition.")
(make-variable-buffer-local 'bee-info-files-list)

;*---------------------------------------------------------------------*/
;*    bee-pop-to-info ...                                              */
;*    -------------------------------------------------------------    */
;*    Pop up a buffer displaying an info file.                         */
;*---------------------------------------------------------------------*/
(defun bee-pop-to-info ()
  (let ((pop-up-frames t))
    (pop-to-buffer "*info*")))

;*---------------------------------------------------------------------*/
;*    bee-info-ref-internal ...                                        */
;*    -------------------------------------------------------------    */
;*    Search in the info documentation for a function (or variable)    */
;*    definition.                                                      */
;*---------------------------------------------------------------------*/
(defun bee-info-ref-internal (var)
  (if (string= var "")
      (progn
	(save-window-excursion
	  (info)
	  (Info-find-node bee-bigloo-info-file "Top"))
	(bee-pop-to-info))
    (let ((found nil))
      (save-window-excursion
	(info)
	(let ((ifiles bee-info-file-list))
	  (while (and (not found) (consp ifiles))
	    (Info-find-node (car ifiles) "Top")
	    (condition-case nil
		(progn
		  (Info-index var)
		  (setq found t))
	      (error (setq ifiles (cdr ifiles)))))))
      (if (not found)
	  (error (format "Can't find documentation for %S" var))
	(bee-pop-to-info)))))

;*---------------------------------------------------------------------*/
;*    bee-info-ref ...                                                 */
;*    -------------------------------------------------------------    */
;*    This function is a wrapper for `bee-info-ref-internal'.          */
;*---------------------------------------------------------------------*/
(defun bee-info-ref (var)
  "Look up a Bigloo function in the Bigloo manual in the Info system.
This command is designed to be used whether you are already in Info or not."
  (interactive (let ((fn (function-at-point))
		     (enable-recursive-minibuffers t)	     
		     val)
		 (setq val (completing-read
			    (format "Look up Bigloo function%s: "
				    (if fn
					(format " (default %s)" fn)
				      ""))
			    obarray 'fboundp t))
		 (list (if (equal val "")
			   fn (intern val)))))
  (bee-info-ref-internal var))

;*---------------------------------------------------------------------*/
;*    bee-info-section ...                                             */
;*    -------------------------------------------------------------    */
;*    Search in the info documentation for a manual section.           */
;*---------------------------------------------------------------------*/
(defun bee-info-section (section)
  "Look up a Bigloo function in the Bigloo manual in the Info system.
This command is designed to be used whether you are already in Info or not."
  (interactive "sSection: ")
  (let ((found nil))
    (save-window-excursion
      (info)
      (let ((ifiles bee-info-file-list))
	(while (and (not found) (consp ifiles))
	  (Info-find-node (car ifiles) "Top")
	  (condition-case nil
	      (progn
		(Info-goto-node section t (car ifiles))
		(setq found t))
	    (error (setq ifiles (cdr ifiles)))))))
    (if (not found)
	(error (format "Can't find section -- %S" section))
      (bee-pop-to-info))))

;*---------------------------------------------------------------------*/
;*    bee-info-sexp ...                                                */
;*    -------------------------------------------------------------    */
;*    This function checks its string argument in order to             */
;*    discover what kind of documentation is required. For instance,   */
;*    this function may decide to print the number, string or          */
;*    application documentation instead of the documentation of a      */
;*    particular pre-defined function.                                 */
;*---------------------------------------------------------------------*/
(defun bee-info-sexp (beg end)
  "Popup an online documentation according to the active region."
  (interactive "r")
  (let ((string (buffer-substring beg end)))
    (cond 
     ((string-match "^;" string)
      ;; this is a comment
      (bee-info-section "Comments"))
     ((string-match "^[0-9.]+$" string)
      ;; this is a number
      (bee-info-section "Numbers"))
     ((string-match "^\"[^\"]*\"" string)
      ;; this is a string
      (bee-info-section "Strings"))
     ((string-match "^((" string)
      ;; this is a computed call
      (bee-info-section "procedure call"))
     ((string-match "['`,]" string)
      ;; this is a quotation
      (bee-info-section "quotation"))
     ((string-match "#[ftFT]$" string)
      ;; booleans
      (bee-info-section "Booleans"))
     ((string-match "#[\\]" string)
      ;; chars
      (bee-info-section "Characters"))
     ((string-match "(\\(\\w+\\)" string)
      ;; an application
      (bee-info-ref-internal (substring string
					(match-beginning 1)
					(match-end 1))))
     (t
      (bee-info-ref-internal string)))))

;*---------------------------------------------------------------------*/
;*    bee-docline ...                                                  */
;*    -------------------------------------------------------------    */
;*    This function calls an online documentation with respect         */
;*    to the active context.                                           */
;*    -------------------------------------------------------------    */
;*    If a region is active, this function print the documentation     */
;*    for the s-expression.                                            */
;*---------------------------------------------------------------------*/
(defun bee-docline ()
  "Popup an online documentation according to context."
  (interactive)
  (if (and (region-active-p) (region-exists-p))
      (bee-info-sexp (region-beginning) (region-end))
    (bee-info-ref-internal (bee-fetch-then-request-identifier (point)
							      "help for: "))))

;*---------------------------------------------------------------------*/
;*    bee-c-docline ...                                                */
;*    -------------------------------------------------------------    */
;*    This function calls an online documentation with respect         */
;*    to the active context.                                           */
;*    -------------------------------------------------------------    */
;*    If a region is active, this function print the documentation     */
;*    for the s-expression.                                            */
;*---------------------------------------------------------------------*/
(defun bee-c-docline ()
  "Popup an online documentation according to context."
  (interactive)
  (let* ((ident (bee-fetch-then-request-c-identifier (point) "help for: "))
	 (Manual-buffer-view-mode 'other)
	 (pop-up-frames t))
    (manual-entry ident bee-c-manual-page-number)))

;*---------------------------------------------------------------------*/
;*    help variables ...                                               */
;*---------------------------------------------------------------------*/
(defvar bee-help-temporary-buffer nil
  "The temporary buffer for expansion.")

(defvar bee-external-help-process '()
  "The external Bee help process")

(defvar bee-external-help-completed nil)
(defun bee-external-help-sentinel (proc msg)
  (setq bee-external-help-completed t)
  (if (equal (substring msg 0 8) "finished")
      (message "Help done ...")
    (progn
      (beep)
      (message "Abnormal end: %s" msg ))))

;*---------------------------------------------------------------------*/
;*    bee-set-help-temporary-buffer ...                                */
;*---------------------------------------------------------------------*/
(defun bee-set-help-temporary-buffer ()
  (if (bufferp bee-help-temporary-buffer)
      (save-excursion
	(set-buffer bee-help-temporary-buffer)
	(toggle-read-only nil)
	(erase-buffer))
    (setq bee-help-temporary-buffer (get-buffer-create "*bee tmp help*"))))

;*---------------------------------------------------------------------*/
;*    bee-help-buffer-internal ...                                     */
;*---------------------------------------------------------------------*/
(defun bee-help-buffer-internal (&optional completion-hook)
  ;; we now help the buffer
  (set-buffer bee-help-temporary-buffer)
  (setq bee-external-help-process
	(start-process "help" (buffer-name) bee-bigloo "-help2"))
  (set-process-sentinel bee-external-help-process
			'bee-external-help-sentinel)
  ;; we wait for completion
  (while (not bee-external-help-completed)
    (sit-for 0.1))
  ;; expansion is completed, we can keep going
  (if completion-hook
      (funcall completion-hook bee-help-temporary-buffer))
  (set-buffer-modified-p nil)
  (toggle-read-only t)
  (let* ((pop-up-frames t)
	 (lines      (+ 3 (count-lines (point-min) (point-max))))
	 (cur-height (frame-height (selected-frame)))
	 (default-frame-alist
	   (if (< lines cur-height)
	       (cons 'height
		     (cons lines
			   (cons 'minibuffer
				 (cons nil
				       default-frame-alist))))
	     (cons 'minibuffer
		   (cons nil
			 default-frame-alist)))))
    (beginning-of-buffer)
    (display-buffer (current-buffer))))

;*---------------------------------------------------------------------*/
;*    bee-compiler-help ...                                            */
;*---------------------------------------------------------------------*/
(defun bee-compiler-help ()
  "Popup an online documentation about the compiler invokation."
  (interactive)
  (setq bee-external-help-completed nil)
  (message "Generating help buffer...")
  (bee-set-help-temporary-buffer)
  (bee-help-buffer-internal))

;*---------------------------------------------------------------------*/
;*    bee-help-buffer-c-internal ...                                   */
;*---------------------------------------------------------------------*/
(defun bee-help-buffer-c-internal ()
  (let ((Manual-buffer-view-mode 'other)
	(pop-up-frames t))
    (manual-entry bee-cc)))
  
;*---------------------------------------------------------------------*/
;*    bee-c-compiler-help ...                                          */
;*---------------------------------------------------------------------*/
(defun bee-c-compiler-help ()
  "Popup an online documentation about the compiler invokation."
  (interactive)
  (setq bee-external-help-completed nil)
  (message "Generating help buffer...")
  (bee-set-help-temporary-buffer)
  (bee-help-buffer-c-internal))

;*---------------------------------------------------------------------*/
;*    doc source keymap                                                */
;*---------------------------------------------------------------------*/
(defvar bee-path-mouse-map (make-sparse-keymap))
(define-key bee-path-mouse-map [(button2)]
  (function bee-path-find))
(defvar bee-deffn-mouse-map (make-sparse-keymap))
(define-key bee-deffn-mouse-map [(button2)]
  (function bee-deffn-find))
(defvar bee-node-mouse-map (make-sparse-keymap))
(define-key bee-node-mouse-map [(button2)]
  (function bee-node-find))

;*---------------------------------------------------------------------*/
;*    bee-fontify-doc-source ...                                       */
;*    -------------------------------------------------------------    */
;*    This function fontify a buffer according to doc sources          */
;*    annotations.                                                     */
;*---------------------------------------------------------------------*/
(defun bee-fontify-doc-source (buffer)
  (interactive "Bbuffer: ")
  (save-excursion
    (set-buffer buffer)
    (beginning-of-buffer)
    (let* ((path-regexp "@path[ \t]+\\([^@]+\\)@")
	   (node-regexp "@node[ \t]+\\([^@]+\\)@")
	   (doc-regexp  "@deffn ?[^@]*@")
	   (union-regexp (concat path-regexp "\\|"
				 node-regexp "\\|"
				 doc-regexp))
	   (path '()))
      (while (re-search-forward union-regexp (point-max) t)
	(goto-char (match-beginning 0))
	(let ((end (match-end 0)))
	  (cond
	   ((re-search-forward path-regexp end t)
	    ;; a path addition
	    (let ((file (buffer-substring (match-beginning 1) (match-end 1))))
	      (if (and (stringp file) (file-exists-p file))
		  (let ((extent (make-extent (match-beginning 0)
					     (match-end 0))))
		    (set-extent-property extent
					 'doc-source file)
		    (set-extent-property extent
					 'mouse-face 'highlight)
		    (set-extent-property extent
					 'keymap bee-path-mouse-map)
		    (setq path (cons file path))))))
	   ((re-search-forward doc-regexp end t)
	    ;; a function definition
	    (let* ((start  (match-beginning 0))
		   (stop   (match-end 0))
		   (extent (make-extent start stop)))
	      (set-extent-property extent 'doc-source path)
	      (set-extent-property extent 'mouse-face 'highlight)
	      (set-extent-property extent 'doc-position (cons start stop))
	      (set-extent-property extent 'keymap bee-deffn-mouse-map)))
	   ((re-search-forward node-regexp end t)
	    ;; a node definition
	    (let ((node (buffer-substring (match-beginning 1) (match-end 1))))
	      (if (stringp node)
		  (let ((extent (make-extent (match-beginning 0)
					     (match-end 0))))
		    (set-extent-property extent
					 'doc-source
					 (cons node path))
		    (set-extent-property extent
					 'mouse-face 'highlight)
		    (set-extent-property extent
					 'keymap bee-node-mouse-map)))))
	   (t
	    (error (format "Illegal match %S" end))))
	  (goto-char end))))))

;*---------------------------------------------------------------------*/
;*    bee-path-find ...                                                */
;*---------------------------------------------------------------------*/
(defun bee-path-find  (event)
  (interactive "e")
  (let* ((extent (event-glyph-extent event))
	 (point  (event-closest-point event))
	 (buffer (event-buffer event))
	 (extent (extent-at point buffer 'doc-source))
	 (file   (extent-property extent 'doc-source)))
    (find-file-other-frame file)))

;*---------------------------------------------------------------------*/
;*    bee-find-in-files ...                                            */
;*---------------------------------------------------------------------*/
(defun bee-find-in-files (pat files)
  (let ((found nil))
    (while (and (not found) (consp files))
      (let* ((cmd (concat "egrep -E -e \"" pat "\" " (car files)))
	     (res (exec-to-string cmd)))
	(if (string= res "")
	    (setq files (cdr files))
	  (setq found (car files)))))
    found))

;*---------------------------------------------------------------------*/
;*    bee-deffn-find ...                                               */
;*---------------------------------------------------------------------*/
(defun bee-deffn-find (event)
  (interactive "e")
  (let* ((extent     (event-glyph-extent event))
	 (point      (event-closest-point event))
	 (buffer     (event-buffer event))
	 (extent     (extent-at point buffer 'doc-source))
	 (files      (extent-property extent 'doc-source))
	 (pos        (extent-property extent 'doc-position))
	 (start      (car pos))
	 (stop       (cdr pos))
	 (doc-regexp "@deffn \\([^@]+\\)@")
	 (id         (save-excursion
		       (goto-char start)
		       (if (re-search-forward doc-regexp stop t)
			   (buffer-substring (match-beginning 1)
					     (match-end 1))
			 (let ((sexp (let ((s (bee-find-toplevel-sexp point)))
				       (if (consp s)
					   s
					 (progn
					   (goto-char point)
					   (if (search-forward "("
							       (point-max)
							       t)
					       (bee-find-toplevel-sexp (point))
					     nil))))))
			   (if (consp sexp)
			       (progn
				 (goto-char (car sexp))
				 (cond
				  ((re-search-forward
				    "(define-\\(generic\\|inline\\)[ \t]+(\\(\\w+\\)"
				    (cdr sexp)
				    t)
				   (buffer-substring (match-beginning 2)
						     (match-end 2)))
				  ((re-search-forward
				    "(define[ \t]+(?\\(\\w+\\)"
				    (cdr sexp)
				    t)
				   (buffer-substring (match-beginning 1)
						     (match-end 1)))
				  (t
				   nil)))))))))
	    (if id
		(let* ((regexp (concat "@deffn[x]?[ \t]*{[^}]+}[ \t]+"
				       id
				       "[\n \t]"))
		       (file (bee-find-in-files regexp files)))
		  (if (stringp file)
		      (let ((buffer (find-file-noselect file)))
			(if (bufferp buffer)
			    (progn
			      (set-buffer buffer)
			      (beginning-of-buffer)
			      (if (re-search-forward regexp (point-max) t)
				  (let ((pop-up-frames t))
				    (goto-char (match-beginning 0))
				    (pop-to-buffer buffer))
				(error "Can't find documentation for %S in %S"
				       id
				       file)))
			  (error "Can't find documentation for %S in %S"
				 id
				 file)))
		    (error "Can't find file %S" file)))
	      (error "Can't find variable name at %S"
		     (buffer-substring start stop)))))

;*---------------------------------------------------------------------*/
;*    bee-node-find ...                                                */
;*---------------------------------------------------------------------*/
(defun bee-node-find (event)
  (interactive "e")
  (let* ((extent (event-glyph-extent event))
	 (point  (event-closest-point event))
	 (buffer (event-buffer event))
	 (extent (extent-at point buffer 'doc-source))
	 (prop   (extent-property extent 'doc-source))
	 (node   (car prop))
	 (files  (cdr prop)))
    (let* ((regexp (concat "@node[ \t]+" node))
	   (file   (bee-find-in-files regexp files)))
      (if (stringp file)
	  (let ((buffer (find-file-noselect file)))
	    (if (bufferp buffer)
		(progn
		  (set-buffer buffer)
		  (beginning-of-buffer)
		  (if (re-search-forward regexp (point-max) t)
		      (let ((pop-up-frames t))
			(goto-char (match-beginning 0))
			(pop-to-buffer buffer))
		    (error "Can't find documentation for node %S in %S"
			   node file)))
	      (error "Can't find buffer for %S" file)))
	(error "Can't find buffer for %S" file)))))
	 
	    
